#!/usr/bin/env node

import { intro, outro, text, spinner, note, isCancel, cancel, select } from '@clack/prompts';
import * as fs from 'fs/promises';
import * as path from 'path';
import colors from 'picocolors';
import { Vercel } from "@vercel/sdk";
import { existsSync } from 'fs';

interface FileTreeEntry {
  name: string;
  type: 'directory' | 'file' | 'symlink' | 'lambda' | 'middleware' | 'invalid';
  uid?: string;
  children?: FileTreeEntry[];
  contentType?: string;
  mode: number;
}

interface FileContent {
  data: string;
}

let vercel: Vercel;

async function getDeployments() {
  const deployments = await vercel.deployments.getDeployments({
    limit: 100,
  });

  return deployments.deployments;
}

async function createDirectory(dirPath: string): Promise<void> {
  await fs.mkdir(dirPath, { recursive: true });
}

async function writeFile(filePath: string, content: string): Promise<void> {
  try {
    const buffer = Buffer.from(content, 'base64');
    const dir = path.dirname(filePath);
    await fs.mkdir(dir, { recursive: true });
    await fs.writeFile(filePath, buffer);
  } catch (error) {
    console.error(`Failed to write file ${filePath}:`, error);
    throw error;
  }
}

async function downloadFile(
  token: string,
  deploymentId: string,
  fileId: string,
  filePath: string,
  s: ReturnType<typeof spinner>
): Promise<void> {
  try {
    const response = await fetch(
      `https://api.vercel.com/v7/deployments/${deploymentId}/files/${fileId}`,
      {
        headers: {
          Authorization: `Bearer ${token}`,
          Accept: '*/*',
        },
      }
    );

    if (!response.ok) {
      throw new Error(`Failed to download file: ${response.statusText}`);
    }

    const { data } = await response.json() as FileContent;
    await writeFile(filePath, data);
    s.message(`Downloaded: ${colors.green(filePath)}`);
  } catch (error) {
    s.message(`Failed: ${colors.red(filePath)}`);
    throw error;
  }
}

async function processEntry(
  token: string,
  deploymentId: string,
  entry: FileTreeEntry,
  currentPath: string,
  s: ReturnType<typeof spinner>
): Promise<void> {
  const entryPath = path.join(currentPath, entry.name);

  if (entry.type === 'directory') {
    await createDirectory(entryPath);
    if (entry.children) {
      for (const child of entry.children) {
        await processEntry(token, deploymentId, child, entryPath, s);
      }
    }
  } else if (entry.type === 'file' && entry.uid) {
    await downloadFile(token, deploymentId, entry.uid, entryPath, s);
  }
}

async function downloadDeployment(
  token: string,
  deploymentId: string,
  outputDir: string
): Promise<void> {
  const s = spinner();
  try {
    s.start('Fetching deployment files');

    await createDirectory(outputDir);

    s.message(`Fetching deployment files for ${colors.green(deploymentId)}`);

    const deployment = await vercel.deployments.getDeployment({
      idOrUrl: deploymentId,
    })

    if (!deployment) {
      s.stop('Deployment not found');
      process.exit(1);
    }

    s.message(`Fetching files for ${colors.green(deployment.name)}`);

    const files = await vercel.deployments.listDeploymentFiles({
      id: deployment.id,
    });

    s.message(`Found ${files.length} files`);

    if (!Array.isArray(files) || files.length === 0) {
      throw new Error('No files found in deployment');
    }

    s.message(`Processing ${files.length} entries`);

    for (const entry of files) {
      try {
        await processEntry(token, deploymentId, entry, outputDir, s);
      } catch (error) {
        console.error(`Failed to process entry ${entry.name}:`, error);
      }
    }

    s.stop('Download completed successfully');
  } catch (error) {
    s.stop('Download failed');
    throw error;
  }
}

async function main() {
  intro(`${colors.cyan('DLVercel - Download the source code of a Vercel deployment.')}`);

  const token = await text({
    message: 'Enter your Vercel token (leave empty to use the TOKEN environment variable)',
    placeholder: 'vercel_token_xxx',
    defaultValue: process.env.TOKEN,
    validate(value) {
      if (value.length === 0 && !process.env.TOKEN) return 'Token is required when TOKEN environment variable is not set';
    },
  });

  if (isCancel(token)) {
    cancel('Operation cancelled');
    process.exit(0);
  }

  const s = spinner();
  s.start('Validating token');

  vercel = new Vercel({
    bearerToken: token,
  });

  try {
    await vercel.user.getAuthUser();

    s.stop('Token validated');
  } catch (error) {
    cancel('The informed token is invalid');
    process.exit(0);
  }

  s.start('Fetching deployments');
  const deployments = await getDeployments();
  s.stop('Deployments fetched');

  if (deployments.length === 0) {
    cancel('No deployments found');
    process.exit(1);
  }

  const selectedDeployment = await select({
    message: 'Select a deployment to download',
    options: deployments.map(deployment => ({
      value: deployment.uid,
      label: `${deployment.name} (${deployment.target || 'preview'}) - ${new Date(deployment.created).toLocaleString()}`,
      hint: deployment.url
    })),
  });

  if (isCancel(selectedDeployment)) {
    cancel('Operation cancelled');
    process.exit(0);
  }

  const outputDir = await text({
    message: 'Where should files be saved?',
    placeholder: './deployment-files',
    initialValue: './deployment-files',
    validate(value) {
      if (value.length === 0) return 'Output directory is required';
      if (existsSync(value)) return 'Directory already exists';
    },
  });

  if (isCancel(outputDir)) {
    cancel('Operation cancelled');
    process.exit(0);
  }

  try {
    await downloadDeployment(token, selectedDeployment, outputDir);

    note(
      `Files downloaded to ${colors.green(path.resolve(outputDir))}`,
      'Success!'
    );
  } catch (error) {
    console.error(error);
  }

  outro('All done! 🎉');
}

main().catch((error) => {
  console.error('Fatal error:', error);
  process.exit(1);
});
